/*
 * Copyright 1995-2008 Sun Microsystems, Inc.  All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

package java.lang;

import java.io.*;
import java.util.Vector;
/*ibm@JSE-8409 start*//*IBM-ios start*/
import sun.security.action.GetPropertyAction;
import java.security.AccessController;
import java.util.StringTokenizer;
/*ibm@JSE-8409 end*//*IBM-ios end*/

/* java.lang.Process subclass in the UNIX environment.
 *
 * @author Mario Wolczko and Ross Knippel.
 * @author Konstantin Kladko (ported to Linux)
 */

final class UNIXProcess extends Process {
    private static final sun.misc.JavaIOFileDescriptorAccess fdAccess
        = sun.misc.SharedSecrets.getJavaIOFileDescriptorAccess();

    private int pid;
    private int exitcode;
    private boolean hasExited;

    private OutputStream stdin_stream;
    private InputStream stdout_stream;
    private InputStream stdout_inner_stream;
    private InputStream stderr_stream;

    /* this is for the reaping thread */
    private native int waitForProcessExit(int pid);

    /**
     * Create a process using fork(2) and exec(2).
     *
     * @param std_fds array of file descriptors.  Indexes 0, 1, and
     *        2 correspond to standard input, standard output and
     *        standard error, respectively.  On input, a value of -1
     *        means to create a pipe to connect child and parent
     *        processes.  On output, a value which is not -1 is the
     *        parent pipe fd corresponding to the pipe which has
     *        been created.  An element of this array is -1 on input
     *        if and only if it is <em>not</em> -1 on output.
     * @return the pid of the subprocess
     */
    private native int forkAndExec(byte[] prog,
                                   byte[] argBlock, int argc,
                                   byte[] envBlock, int envc,
                                   byte[] dir,
                                   int[] std_fds,
                                   boolean redirectErrorStream)
        throws IOException;

    /* In the process constructor we wait on this gate until the process    */
    /* has been created. Then we return from the constructor.               */
    /* fork() is called by the same thread which later waits for the process */
    /* to terminate */

    private static class Gate {

        private boolean exited = false;
        private IOException savedException;

        synchronized void exit() { /* Opens the gate */
           exited = true;
           this.notify();
        }

        synchronized void waitForExit() { /* wait until the gate is open */
            boolean interrupted = false;
            while (!exited) {
                try {
                    this.wait();
                } catch (InterruptedException e) {
                    interrupted = true;
                }
            }
            if (interrupted) {
                Thread.currentThread().interrupt();
            }
        }

        void setException (IOException e) {
            savedException = e;
        }

        IOException getException() {
            return savedException;
        }
    }

    UNIXProcess(final byte[] prog,
                final byte[] argBlock, final int argc,
                final byte[] envBlock, final int envc,
                final byte[] dir,
                final int[] std_fds,
                final boolean redirectErrorStream)
    throws IOException {

        final Gate gate = new Gate();
        /*
         * For each subprocess forked a corresponding reaper thread
         * is started.  That thread is the only thread which waits
         * for the subprocess to terminate and it doesn't hold any
         * locks while doing so.  This design allows waitFor() and
         * exitStatus() to be safely executed in parallel (and they
         * need no native code).
         */

        java.security.AccessController.doPrivileged(
        new java.security.PrivilegedAction<Void>() {
        public Void run() {
            Thread t = new Thread("process reaper") {
                    public void run() {
                        try {
                        	/*ibm@JSE-8409 start*//*IBM-ios start*/
                        	if(os400 == 1) {
                        		       
                            String filename = new String(prog,platformEncoding);
                            final byte fullprog[] = fullPath(prog);
                            filename = new String(fullprog,platformEncoding);
                            pid = forkAndExec(fullprog,
                                              argBlock, argc,
                                              envBlock, envc,
                                              dir,
                                              std_fds,
                                              redirectErrorStream);
                            }
                           /*ibm@JSE-8409 end*//*IBM-ios end*/
                          else {
                            pid = forkAndExec(prog,
                                              argBlock, argc,
                                              envBlock, envc,
                                              dir,
                                              std_fds,
                                              redirectErrorStream);
                          	
                          }
                        } catch (IOException e) {
                            gate.setException(e); /*remember to rethrow later*/
                            gate.exit();
                            return;
                        }
                        java.security.AccessController.doPrivileged(
                    new java.security.PrivilegedAction<Void>() {
                    public Void run() {
                        if (std_fds[0] == -1)
                            stdin_stream = ProcessBuilder.NullOutputStream.INSTANCE;
                        else {
                            FileDescriptor stdin_fd = new FileDescriptor();
                            fdAccess.set(stdin_fd, std_fds[0]);
                            stdin_stream = new BufferedOutputStream(
                                new FileOutputStream(stdin_fd));
                        }

                        if (std_fds[1] == -1)
                            stdout_stream = ProcessBuilder.NullInputStream.INSTANCE;
                        else {
                            FileDescriptor stdout_fd = new FileDescriptor();
                            fdAccess.set(stdout_fd, std_fds[1]);
			    stdout_inner_stream = new DeferredCloseInputStream(stdout_fd);
                            stdout_stream = new BufferedInputStream(stdout_inner_stream);
                        }

                        if (std_fds[2] == -1)
                            stderr_stream = ProcessBuilder.NullInputStream.INSTANCE;
                        else {
                            FileDescriptor stderr_fd = new FileDescriptor();
                            fdAccess.set(stderr_fd, std_fds[2]);
                            stderr_stream = new DeferredCloseInputStream(stderr_fd);
                        }

                        return null; }});
                        gate.exit(); /* exit from constructor */
                        int res = waitForProcessExit(pid);
                        synchronized (UNIXProcess.this) {
                            hasExited = true;
                            exitcode = res;
                            UNIXProcess.this.notifyAll();
                        }
                    }
                };
                t.setDaemon(true);
                t.start();
                return null; }});
        gate.waitForExit();
        IOException e = gate.getException();
        if (e != null)
            throw new IOException(e.toString());
    }

    public OutputStream getOutputStream() {
        return stdin_stream;
    }

    public InputStream getInputStream() {
        return stdout_stream;
    }

    public InputStream getErrorStream() {
        return stderr_stream;
    }

    public synchronized int waitFor() throws InterruptedException {
        while (!hasExited) {
            wait();
        }
        return exitcode;
    }

    public synchronized int exitValue() {
        if (!hasExited) {
            throw new IllegalThreadStateException("process hasn't exited");
        }
        return exitcode;
    }

    private static native void destroyProcess(int pid);
    public void destroy() {
        // There is a risk that pid will be recycled, causing us to
        // kill the wrong process!  So we only terminate processes
        // that appear to still be running.  Even with this check,
        // there is an unavoidable race condition here, but the window
        // is very small, and OSes try hard to not recycle pids too
        // soon, so this is quite safe.
        synchronized (this) {
            if (!hasExited)
                destroyProcess(pid);
        }
        try {
            stdin_stream.close();
            stdout_stream.close();
            stderr_stream.close();
        } catch (IOException e) {
            // ignore
        }
    }
/*ibm@JSE-8409 start*//*IBM-ios start*/
    private native int statExecutable(byte []filename_bytes);
     
    public byte[] fullPath(byte []filename_bytes) throws IOException {
    	
        String sep = (String)AccessController.doPrivileged           /*ibm@87093*/
                         (new GetPropertyAction("file.separator"));  /*ibm@87093*/
        String PATH = ProcessEnvironment.getenv("PATH");
        String ps = (String)AccessController.doPrivileged            /*ibm@87093*/
                         (new GetPropertyAction("path.separator"));  /*ibm@87093*/
        String fullfilename = null;
        byte fullfilename_bytes[] = null;

        String filename;
        /* unterminatedFilename contains filename without the null character */
        String unterminatedFilename;                             /*ibm@133117*/

        try {
            filename = new String(filename_bytes,platformEncoding);
        } catch (UnsupportedEncodingException e) {
            filename = new String(filename_bytes);
        }
        unterminatedFilename = filename.replace("\u0000","");    /*ibm@133117*/


        /* If the filename we want to exec has any slashes in it then  */
        /* we shouldn't do a path search, as in /foo ./foo or foo/bar. */
        if ((filename.indexOf(sep) < 0) && PATH != null) {
        /*ibm@138341 start*/
        if (os400 == 1) {
        /* Prepend java.home to the PATH for command resolution. i5/OS needs */
        /* to check the java.home location before anything in the PATH.      */
        String java_home = (String)AccessController.doPrivileged
                          (new GetPropertyAction("java.home"));
        if (java_home != null) {
          StringBuffer PREPEND = new StringBuffer();
          PREPEND.append(java_home + sep + ".." + sep + "bin" + ps);
          PREPEND.append(java_home + sep + "bin" + ps);
          PATH = PREPEND.toString() + PATH;
          if (ProcessEnvironment.getenv("QIBM_QYJ9NTVUTL_DEBUG") != null) {
              System.out.println("UNIXProcess.java input pprog=" + filename);
              System.out.println("UNIXProcess.java PATH checked=" + PATH);
           }
        }
      }
      /*ibm@138341 end*/

       StringTokenizer st = new StringTokenizer(PATH,ps);
       boolean foundit = false;
       while (st.hasMoreTokens() && !foundit) {
           fullfilename = st.nextToken() + sep + filename;
           try {
               fullfilename_bytes =
                   fullfilename.getBytes(platformEncoding);
           } catch (UnsupportedEncodingException e) {
               fullfilename_bytes = fullfilename.getBytes();
           }
           if ( statExecutable(fullfilename_bytes) == 0) {
               foundit = true;
           }
     }
       if (!foundit) throw new IOException(unterminatedFilename +
                         ": not found");       /*ibm@133117*/

     } else {
            int status = statExecutable(filename_bytes);
            switch (status) {
                case -1: throw new IOException(unterminatedFilename +
                                         ": not found");        /*ibm@133117*/
                case -2: throw new IOException(unterminatedFilename+
                                         ": cannot execute");   /*ibm@133117*/
            }
            fullfilename_bytes = filename_bytes;
        }

        return fullfilename_bytes;
    }
   

/*ibm@JSE-8409 end*//*IBM-ios end*/

// A FileInputStream that supports the deferment of the actual close
    // operation until the last pending I/O operation on the stream has
    // finished.  This is required on Solaris because we must close the stdin
    // and stdout streams in the destroy method in order to reclaim the
    // underlying file descriptors.  Doing so, however, causes any thread
    // currently blocked in a read on one of those streams to receive an
    // IOException("Bad file number"), which is incompatible with historical
    // behavior.  By deferring the close we allow any pending reads to see -1
    // (EOF) as they did before.
    //
    private static class DeferredCloseInputStream
        extends FileInputStream
    {

        private DeferredCloseInputStream(FileDescriptor fd) {
            super(fd);
        }

        private class InterruptLock {}
        private Object lock = new InterruptLock();      // For the following fields
        private boolean closePending = false;
        private int useCount = 0;
        private InputStream streamToClose;

        private void raise() {
            synchronized (lock) {
                useCount++;
            }
        }

        private void lower() throws IOException {
            synchronized (lock) {
                useCount--;
                if (useCount == 0 && closePending) {
                    streamToClose.close();
                }
            }
        }

        // stc is the actual stream to be closed; it might be this object, or
        // it might be an upstream object for which this object is downstream.
        //
        private void closeDeferred(InputStream stc) throws IOException {
            synchronized (lock) {
                if (useCount == 0) {
                    stc.close();
                } else {
                    closePending = true;
                    streamToClose = stc;
                }
            }
        }

        public void close() throws IOException {
            synchronized (lock) {
                useCount = 0;
                closePending = false;
            }
            super.close();
        }

        public int read() throws IOException {
            raise();
            try {
                return super.read();
            } finally {
                lower();
            }
        }

        public int read(byte[] b) throws IOException {
            raise();
            try {
                return super.read(b);
            } finally {
                lower();
            }
        }

        public int read(byte[] b, int off, int len) throws IOException {
            raise();
            try {
                return super.read(b, off, len);
            } finally {
                lower();
            }
        }

        public long skip(long n) throws IOException {
            raise();
            try {
                return super.skip(n);
            } finally {
                lower();
            }
        }

        public int available() throws IOException {
            raise();
            try {
                return super.available();
            } finally {
                lower();
            }
        }

    }


    /* This routine initializes JNI field offsets for the class */
    private static native void initIDs();
    
    /*ibm@JSE-8409 start*//*IBM-ios start*/
    static String platformEncoding;
    private static int os400 = -1;    
    private static void initiOsFlag() {
       if (os400 < 0) {
		    String os_name = (String)AccessController.doPrivileged
		                          (new GetPropertyAction("os.name"));
		    os400 = ((os_name != null) && os_name.equals("OS/400")) ? 1 : 0;
		   
        platformEncoding = (String)AccessController.doPrivileged
                               (new GetPropertyAction("ibm.system.encoding"));
        if (platformEncoding == null) {
            platformEncoding = (String)AccessController.doPrivileged
                                   (new GetPropertyAction("console.encoding"));
        }
        if (platformEncoding == null) {
            platformEncoding = (String)AccessController.doPrivileged
                                   (new GetPropertyAction("file.encoding"));
                                  }
	    }
    }
    /*ibm@JSE-8409 end*//*IBM-ios end*/
    static {
        initIDs();
        initiOsFlag();        /*ibm@JSE-8409 *//*IBM-ios_bringup*/
    }
}
